@@ -143,20 +143,11 @@ on_window1_realize                     (GtkWidget       *widget,
   
   XSetWMProtocols (GDK_WINDOW_XDISPLAY (widget->window),
 		   GDK_WINDOW_XWINDOW (widget->window), wm_window_protocols, 3);
+  gdk_error_trap_pop ();
   
   if (is_dock) {
-    gint w, h;
-    gtk_window_get_size (GTK_WINDOW (widget), &w, &h);
-
-    atom_window_type = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE");
-    atom_type_dock[0] = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE_DOCK");
-    XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
-		     GDK_WINDOW_XWINDOW (widget->window), 
-		     atom_window_type,
-		     XA_ATOM, 32, PropModeReplace,
-		     (guchar *) &atom_type_dock, 1);
+	  gok_main_set_wm_dock (TRUE);
   }
-  gdk_error_trap_pop ();
 
   gtk_window_stick (GTK_WINDOW (widget));
 
