@@ -705,7 +705,7 @@ GokDockType gok_data_get_dock_type ()
  * gok_data_set_dock_type:
  * @val: a GokDockType indicating whether GOK's main keyboard should be a "DOCK" window.
  */
-void gok_data_set_is_dock (GokDockType val)
+void gok_data_set_dock_type (GokDockType val)
 {
 	m_eDockType = val;
 	gchar *typestring;
