@@ -665,6 +665,11 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		pKey->Type = KEYTYPE_BRANCHEDITTEXT;
 		pKey->Style = KEYSTYLE_BRANCHEDITTEXT;
 	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchSettings") == 0)
+	{
+		pKey->Type = KEYTYPE_BRANCH;
+		pKey->Style = KEYSTYLE_SETTINGS;
+	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"settings") == 0)
 	{
 		pKey->Type = KEYTYPE_SETTINGS;
@@ -680,6 +685,20 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		pKey->Type = KEYTYPE_POINTERCONTROL;
 		pKey->Style = KEYSTYLE_POINTERCONTROL;
 	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"dock") == 0)
+	{
+		pKey->Type = KEYTYPE_DOCK;
+		pKey->Style = KEYSTYLE_DOCK;
+		pKey->pGeneral = g_new0 (GokKeyboardDirection, 1);
+		pStringAttributeValue = 
+			xmlGetProp (pNode, (const xmlChar *) "dir");
+		if (pStringAttributeValue != NULL) 
+		{
+			* (GokKeyboardDirection *) pKey->pGeneral = 
+				gok_keyboard_parse_direction (pStringAttributeValue);
+			g_free (pStringAttributeValue);
+		}
+	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"move-resize") == 0)
 	{
 		pKey->Type = KEYTYPE_MOVERESIZE;
@@ -694,6 +713,16 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 			g_free (pStringAttributeValue);
 		}
 	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"help") == 0)
+	{
+		pKey->Type = KEYTYPE_HELP;
+		pKey->Style = KEYSTYLE_HELP;
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"about") == 0)
+	{
+		pKey->Type = KEYTYPE_ABOUT;
+		pKey->Style = KEYSTYLE_ABOUT;
+	}
 
 	/* branch target */
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "target");
@@ -1358,6 +1387,7 @@ void gok_key_set_button_name (GokKey* pKey)
 	
 	switch (pKey->Style)
 	{
+	case KEYSTYLE_DOCK:
 	case KEYSTYLE_POINTERCONTROL:
 	case KEYSTYLE_NORMAL:
 		strcpy (styleButton, "StyleButtonNormal");
@@ -1411,20 +1441,23 @@ void gok_key_set_button_name (GokKey* pKey)
 		strcpy (styleText, "StyleTextNormal");
 		break;
 		
+	case KEYSTYLE_ABOUT: 
 	case KEYSTYLE_BRANCHGUI:
 		strcpy (styleButton, "StyleButtonBranchGUI");
 		strcpy (styleText, "StyleTextNormal");
 		break;
 
 	case KEYSTYLE_BRANCHGUIACTIONS:
+	case KEYSTYLE_HYPERLINK:
 	case KEYSTYLE_EDITABLETEXTACTION:
 		strcpy (styleButton, "StyleButtonBranchGuiActions");
 		strcpy (styleText, "StyleTextNormal");
 		break;
 		
 	case KEYSTYLE_BRANCHALPHABET:
+	case KEYSTYLE_BRANCHTEXT:
 		strcpy (styleButton, "StyleButtonBranchAlphabet");
-		strcpy (styleText, "StyleTextNormal");
+		strcpy (styleText, "StyleTextBranchAlphabet");
 		break;
 		
 	case KEYSTYLE_SETTINGS:
@@ -1432,6 +1465,11 @@ void gok_key_set_button_name (GokKey* pKey)
 		strcpy (styleText, "StyleTextNormal");
 		break;
 					
+	case KEYSTYLE_HELP:
+		strcpy (styleButton, "StyleButtonHelp");
+		strcpy (styleText, "StyleTextHelp");
+		break;
+					
 	case KEYSTYLE_WORDCOMPLETE:
 		strcpy (styleButton, "StyleButtonWordComplete");
 		strcpy (styleText, "StyleTextWordComplete");
