@@ -25,6 +25,8 @@
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
+#define XK_LATIN1
+#include <X11/keysymdef.h> /* ugh, needed by link following hack */
 #include <assert.h>
 #include <gdk/gdk.h>
 #include <math.h>
@@ -52,6 +54,8 @@
 #define NUM_INTERESTING_ROLES 5
 #define GOK_KEYBOARD_MAX_ROWS 24
 
+#define CORE_KEYBOARD "Keyboard"
+
 /* if this flag is set then ignore the resize event (because we generated it) */
 static gint m_bIgnoreResizeEvent;
 
@@ -389,7 +393,7 @@ gok_keyboard_get_core ()
 	GokKeyboard *pKeyboard = gok_keyboard_new ();
 	GokKey *pKey;
 	gok_log_enter();
-	gok_keyboard_set_name (pKeyboard, _("Keyboard"));
+	gok_keyboard_set_name (pKeyboard, _(CORE_KEYBOARD));
 	pKeyboard->bRequiresLayout = FALSE;
 	pKeyboard->bLaidOut = TRUE;
 	pKeyboard->bSupportsWordCompletion = TRUE;
@@ -576,6 +580,20 @@ gboolean gok_keyboard_add_keys (GokKeyboard* pKeyboard, xmlDoc* pDoc)
 		pKeyboard->LayoutType = atoi ((char *)pStringAttributeValue);
 	}
 
+	/* get the expansion policy of the keyboard (this attribute is optional) */
+	pStringAttributeValue = xmlGetProp (pNodeKeyboard, (const xmlChar *) "expand");
+	if (pStringAttributeValue != NULL)
+	{
+		if (xmlStrcmp (pStringAttributeValue, (const xmlChar *) "never") == 0)
+		{
+			pKeyboard->expand = GOK_EXPAND_NEVER;
+		}
+		else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *) "always") == 0)
+		{
+			pKeyboard->expand = GOK_EXPAND_ALWAYS;
+		}
+	}
+
 	/* go through all the keys */
 	pKeyPrevious = NULL;
 	pNodeKey = pNodeKeyboard->xmlChildrenNode;
@@ -722,9 +740,10 @@ GokKeyboard* gok_keyboard_new ()
 	pGokKeyboardNew->bSupportsCommandPrediction = FALSE;
 	pGokKeyboardNew->bWordCompletionKeysAdded = FALSE;
 	pGokKeyboardNew->bCommandPredictionKeysAdded = FALSE;
+	pGokKeyboardNew->expand = GOK_EXPAND_SOMETIMES;
 	pGokKeyboardNew->pAccessible = NULL;	
 	pGokKeyboardNew->pNodeAccessible = NULL;	
-	pGokKeyboardNew->Role = SPI_ROLE_MENU;
+	pGokKeyboardNew->search_role = SPI_ROLE_MENU;
 	
 	m_NumberOfKeyboards++;
 	
@@ -1234,6 +1253,7 @@ gboolean gok_keyboard_display (GokKeyboard* pKeyboard, GokKeyboard* pKeyboardCur
 	GokButton* pGokButton;
 	GtkButton* pButton;
 	gboolean bKeySizeChanged = FALSE;
+	gboolean expand = FALSE;
 
 	g_assert (pKeyboard != NULL);
 	g_assert (pWindowMain != NULL);
@@ -1390,10 +1410,13 @@ gboolean gok_keyboard_display (GokKeyboard* pKeyboard, GokKeyboard* pKeyboardCur
 		widthMax = gdk_screen_width();
 		heightMax = gdk_screen_height();
 	}
-	
+
+	expand = ((pKeyboard->expand == GOK_EXPAND_ALWAYS) || 
+		  (gok_data_get_expand () && (pKeyboard->expand != GOK_EXPAND_NEVER)));
+
 	if (((widthWindow + frameX) > widthMax) || 
-	    (pKeyboard->LayoutType == KEYBOARD_LAYOUT_FITWINDOW) ||
-	    gok_data_get_expand ())
+	    (pKeyboard->LayoutType == KEYBOARD_LAYOUT_FITWINDOW) || expand)
+
 	{
 		/* change the key width (for this keyboard) to fit within the screen */
 		widthKeyTemp = gok_keyboard_get_keywidth_for_window (widthMax - (frameX * 2), pKeyboard);
@@ -1423,7 +1446,7 @@ gboolean gok_keyboard_display (GokKeyboard* pKeyboard, GokKeyboard* pKeyboardCur
 	       gok_keyboard_calculate_font_size (pKeyboard);
 
 	/* resize the window to hold all the keys */
-	gok_main_resize_window (pWindowMain, widthWindow, heightWindow);
+	gok_main_resize_window (pWindowMain, pKeyboard, widthWindow, heightWindow);
 
 	/* position and resize all the buttons */
 	gok_keyboard_position_keys (pKeyboard, pWindowMain);
@@ -1531,20 +1554,31 @@ gboolean gok_keyboard_branch_byKey (GokKey* pKey)
 
 		case KEYTYPE_BRANCHMENUS:
 			gok_log("branch back SPI_ROLE_MENU");		
-			return gok_keyboard_branch_gui(pKey->pNodeAccessible, SPI_ROLE_MENU); 
+			return gok_keyboard_branch_gui(pKey->pNodeAccessible, 
+						       GOK_SPY_SEARCH_MENU,
+						       SPI_ROLE_MENU); 
 			break;
 
 		case KEYTYPE_BRANCHMENUITEMS:
 			gok_log("branch gui SPI_ROLE_MENU_ITEM");		
-			return gok_keyboard_branch_gui(pKey->pNodeAccessible, SPI_ROLE_MENU_ITEM);
+			return gok_keyboard_branch_gui(pKey->pNodeAccessible, 
+						       GOK_SPY_SEARCH_ROLE, 
+						       /* does this work for check menu items too? */
+						       SPI_ROLE_MENU_ITEM);
 			break;
 			
 		case KEYTYPE_BRANCHTOOLBARS:
+			gok_log("branch gui SPI_ROLE_PUSH_BUTTON");		
+			return gok_keyboard_branch_gui(pKey->pNodeAccessible, 
+						       GOK_SPY_SEARCH_ROLE,
+						       SPI_ROLE_PUSH_BUTTON); 
+			break;
 		case KEYTYPE_BRANCHGUI:
 			gok_log("branch gui SPI_ROLE_PUSH_BUTTON");		
-			return gok_keyboard_branch_gui(pKey->pNodeAccessible, SPI_ROLE_PUSH_BUTTON); /* for now test buttons - abstract out later */
-			break;
-			
+			return gok_keyboard_branch_gui(pKey->pNodeAccessible, 
+						       GOK_SPY_SEARCH_UI,
+						       SPI_ROLE_PUSH_BUTTON); 
+			break;			
 		case KEYTYPE_BRANCHGUIACTIONS:
 			gok_log("branch gui_actions");
 			is_branched = gok_keyboard_branch_gui_actions(pKey->pNodeAccessible);
@@ -1562,7 +1596,9 @@ gboolean gok_keyboard_branch_byKey (GokKey* pKey)
 			gok_log("branch edittext");		
 			return gok_keyboard_branch_edittext(); 
 			break;
-
+	        case KEYTYPE_BRANCHHYPERTEXT:
+			fprintf (stderr, "HYPERTEXT!\n");
+                        break;
 		default:
 			/* should not be here */
 			gok_log ("Unknown branch type");
@@ -1571,6 +1607,16 @@ gboolean gok_keyboard_branch_byKey (GokKey* pKey)
 	return FALSE;
 }
 
+
+static KeyStyles 
+gok_style_if_enabled (AccessibleStateSet *states, KeyStyles style) 
+{
+	if (AccessibleStateSet_contains (states, SPI_STATE_ENABLED))
+		return style;
+	else
+		return KEYSTYLE_INSENSITIVE;
+}
+
 /**
 * gok_keyboard_update_dynamic
 * @pKeyboard: Pointer to the keyboard that gets updated.
@@ -1582,7 +1628,7 @@ gboolean gok_keyboard_branch_byKey (GokKey* pKey)
 gboolean gok_keyboard_update_dynamic (GokKeyboard* pKeyboard)
 {
 	AccessibleNode* pNodeAccessible;
-	AccessibleNode* pNodeAccessibleCount;
+	AccessibleNode* pNodeAccessibleNext;
 	AccessibleStateSet *pStateSet;
 	GokKey* pKey;
 	GokKey* pKeyTemp;
@@ -1618,20 +1664,23 @@ gboolean gok_keyboard_update_dynamic (GokKeyboard* pKeyboard)
 	pKeyboard->pKeyFirst = NULL;
 
 	
-/*
-	if (pKeyboard->Role == GOKSPY_ROLE_APPLICATION)
+#if REINSTATE_APPLICATIONS
+	if (pKeyboard->search_role == GOKSPY_ROLE_APPLICATION)
 	{
 		pNodeAccessible = gok_spy_get_applications();
 	}
 	else
 	{
-*/
-gok_log("calling get list with accessible [%d] and role [%s]", pKeyboard->pAccessible, AccessibleRole_getName(pKeyboard->Role));
-		/* create new keys for the keyboard */
-		pNodeAccessible = gok_spy_get_list (pKeyboard->pAccessible, pKeyboard->Role);
-/*	}
-*/
-	
+#endif
+	gok_log("calling get list with accessible [%d] and role [%s]", 
+		pKeyboard->pAccessible, AccessibleRole_getName(pKeyboard->search_role));
+	/* create new keys for the keyboard */
+
+
+	/* TODO: REFACTOR: enum KeyType (keyboard role) */
+	pNodeAccessible = 
+		gok_spy_get_list (pKeyboard->pAccessible, pKeyboard->search_type, pKeyboard->search_role);
+		
 	if (pNodeAccessible == NULL)
 	{
 		gok_log_x ("Warning: pNodeAccessible is NULL!");
@@ -1660,19 +1709,19 @@ gok_log("calling get list with accessible [%d] and role [%s]", pKeyboard->pAcces
 	pKeyboard->bLaidOut = FALSE;
 
 	column = 1;
-	pNodeAccessibleCount = pNodeAccessible;
+	pNodeAccessibleNext = pNodeAccessible;
 
-	while (pNodeAccessibleCount != NULL)
+	while (pNodeAccessibleNext != NULL)
 	{
 		pKey = gok_key_new (pKeyPrevious, NULL, pKeyboard);
 		pKeyPrevious = pKey;
 		
 		pKey->Style = KEYSTYLE_GENERALDYNAMIC;
 		
-		gok_log("node has role: [%s]",Accessible_getRoleName(pNodeAccessibleCount->paccessible));
+		gok_log("node has role: [%s]",Accessible_getRoleName(pNodeAccessibleNext->paccessible));
 		
-		pStateSet = Accessible_getStateSet (pNodeAccessibleCount->paccessible);
-		switch (Accessible_getRole(pNodeAccessibleCount->paccessible))
+		pStateSet = Accessible_getStateSet (pNodeAccessibleNext->paccessible);
+		switch (Accessible_getRole(pNodeAccessibleNext->paccessible))
 		{
 			/*
 			  case SPI_ROLE_INVALID:
@@ -1683,7 +1732,7 @@ gok_log("calling get list with accessible [%d] and role [%s]", pKeyboard->pAcces
 		case SPI_ROLE_MENU:
 			gok_log("setting key type BRANCHMENUITEMS");
 			pKey->Type = KEYTYPE_BRANCHMENUITEMS;
-			pKey->Style = KEYSTYLE_BRANCHMENUITEMS;
+			pKey->Style = gok_style_if_enabled (pStateSet, KEYSTYLE_BRANCHMENUITEMS);
 			break;
 			/* add cases here	
 			   case SPI_ROLE_XXX:
@@ -1693,35 +1742,66 @@ gok_log("calling get list with accessible [%d] and role [%s]", pKeyboard->pAcces
 		case SPI_ROLE_CHECK_BOX:
 		case SPI_ROLE_CHECK_MENU_ITEM:
 			pKey->Type = KEYTYPE_BRANCHGUIACTIONS;
-			pKey->Style = KEYSTYLE_BRANCHGUIACTIONS;
+			pKey->Style = gok_style_if_enabled (pStateSet, KEYSTYLE_BRANCHGUIACTIONS);
 			is_active = AccessibleStateSet_contains (pStateSet, 
 								 SPI_STATE_CHECKED);
 			pKey->ComponentState.active = is_active;
 			break;
+		case SPI_ROLE_TEXT:
+			/* should only be in the list if it's editable...*/
+			if (pNodeAccessible->is_link) {
+			      pKey->Type = KEYTYPE_HYPERLINK;
+			      pKey->Style = KEYSTYLE_HYPERLINK;
+			      is_active = TRUE;
+			}
+			else {
+			      pKey->Type = KEYTYPE_BRANCHTEXT;
+			      if (AccessibleStateSet_contains (pStateSet, SPI_STATE_EDITABLE)) { 
+				      pKey->Style = gok_style_if_enabled (pStateSet, KEYSTYLE_BRANCHTEXT);
+				      is_active = TRUE;
+			      }
+			      else {
+				      is_active = FALSE;
+				      pKey->Style = KEYSTYLE_INSENSITIVE;
+			      }
+			}
+			pKey->ComponentState.active = is_active;
+			break;
+		case SPI_ROLE_HTML_CONTAINER: /* TODO: check Hypertext interface instead */
+			pKey->Type = KEYTYPE_BRANCHHYPERTEXT;
+			pKey->Style = KEYSTYLE_BRANCH; /* reuse 'normal' branch style for now. */
+			is_active = AccessibleStateSet_contains (pStateSet, 
+								 SPI_STATE_SENSITIVE);
+			pKey->ComponentState.active = is_active;
+			break;
 		default:
-			/* this might be a dangerous catch all... */
-			gok_log("setting key type BRANCHGUIACTIONS");
-			pKey->Type = KEYTYPE_BRANCHGUIACTIONS;
-			pKey->Style = KEYSTYLE_BRANCHGUIACTIONS;
+			if (Accessible_isHypertext (pNodeAccessibleNext->paccessible)) {
+				pKey->Type = KEYTYPE_BRANCHHYPERTEXT;
+				pKey->Style = KEYSTYLE_BRANCH; /* reuse 'normal' branch style for now. */
+				is_active = TRUE; /* FIXME, might not be enabled? */
+				pKey->ComponentState.active = is_active;
+			}
+			else {
+				/* this might be a dangerous catch all... */
+				gok_log("setting key type BRANCHGUIACTIONS");
+				pKey->Type = KEYTYPE_BRANCHGUIACTIONS;
+				pKey->Style = gok_style_if_enabled (pStateSet, KEYSTYLE_BRANCHGUIACTIONS);
+			}
 			break;
 		}
 
-		if (!AccessibleStateSet_contains (pStateSet, SPI_STATE_ENABLED) ) 
-		{
-			pKey->Style = KEYSTYLE_INSENSITIVE;
-		}
 		AccessibleStateSet_unref (pStateSet);
 		pKey->Top = 0;
 		pKey->Bottom = 1;
 		pKey->Left = column;
 		pKey->Right = column + 1;
-		pKey->pNodeAccessible = pNodeAccessibleCount;
+		pKey->pNodeAccessible = pNodeAccessibleNext;
 		
-		gok_log("adding label %s to dynamic key",pNodeAccessibleCount->pname);
-		gok_key_add_label (pKey, pNodeAccessibleCount->pname, NULL);
+		gok_log("adding label %s to dynamic key",pNodeAccessibleNext->pname);
+		gok_key_add_label (pKey, pNodeAccessibleNext->pname, NULL);
 		
 		column++;
-		pNodeAccessibleCount = pNodeAccessibleCount->pnext;
+		pNodeAccessibleNext = pNodeAccessibleNext->pnext;
 	}	
 
 	gok_log_leave();
@@ -1731,13 +1811,15 @@ gok_log("calling get list with accessible [%d] and role [%s]", pKeyboard->pAcces
 /**
 * gok_keyboard_branch_gui
 * @pNodeAccessible: Pointer to the accessible node (parent object).
-* @role: Type of dynamic keyboard to branch to (the role of the children).
+* @type: Type of dynamic keyboard to branch to (the role of the children).
+* @role: role to match, if type is #GOK_SPY_SEARCH_ROLE.
 *
 * Displays the generic gui keyboard - currently used for widgets inside windowish things
 *
 * Returns: TRUE if the keyboard was displayed, FALSE if not.
 **/
-gboolean gok_keyboard_branch_gui (AccessibleNode* pNodeAccessible, AccessibleRole role)
+gboolean gok_keyboard_branch_gui (AccessibleNode* pNodeAccessible, 
+				  GokSpySearchType type, AccessibleRole role)
 {
 	Accessible* pAccessibleRoot;
 	GokKeyboard* pKeyboard;
@@ -1747,26 +1829,23 @@ gboolean gok_keyboard_branch_gui (AccessibleNode* pNodeAccessible, AccessibleRol
 	
 	pAccessibleRoot = NULL;
 
-	if (role != GOKSPY_ROLE_APPLICATION)
+	if (pNodeAccessible != NULL)
 	{
-		if (pNodeAccessible != NULL)
-		{
-			gok_log("using accessible from key as root accessible");
-			pAccessibleRoot = pNodeAccessible->paccessible;
-		}
+		gok_log("using accessible from key as root accessible");
+		pAccessibleRoot = pNodeAccessible->paccessible;
+	}
+	
+	if (pAccessibleRoot == NULL)
+	{
+		gok_log("using accessible for the foregound application as root accessible");
+		/* get the accessible interface for the foregound application */
+		pAccessibleRoot = gok_main_get_foreground_window_accessible();
 		
 		if (pAccessibleRoot == NULL)
 		{
-			gok_log("using accessible for the foregound application as root accessible");
-			/* get the accessible interface for the foregound application */
-			pAccessibleRoot = gok_main_get_foreground_window_accessible();
-			
-			if (pAccessibleRoot == NULL)
-			{
-				gok_log_x ("Warning: Can't create gui keyboard because foreground accessible is NULL!");
-				gok_log_leave();
-				return FALSE;
-			}
+			gok_log_x ("Warning: Can't create gui keyboard because foreground accessible is NULL!");
+			gok_log_leave();
+			return FALSE;
 		}
 	}
 
@@ -1784,8 +1863,6 @@ gboolean gok_keyboard_branch_gui (AccessibleNode* pNodeAccessible, AccessibleRol
 	/* store the accessible pointer on the keyboard */
 	gok_keyboard_set_accessible(pKeyboard, pAccessibleRoot);
 
-	pKeyboard->Role = role;
-		
 	/* add the new keyboard to the list of keyboards (at the end)*/
 	pKeyboardTemp = gok_main_get_first_keyboard();
 	g_assert (pKeyboardTemp != NULL);
@@ -1795,37 +1872,51 @@ gboolean gok_keyboard_branch_gui (AccessibleNode* pNodeAccessible, AccessibleRol
 	}
 	pKeyboardTemp->pKeyboardNext = pKeyboard;
 	pKeyboard->pKeyboardPrevious = pKeyboardTemp;
-	
+	pKeyboard->search_type = type;
+	pKeyboard->search_role = role;
+
 	/* set the name and type of the keyboard */
-	switch (role)
+	switch (type)
 	{
+	case GOK_SPY_SEARCH_UI:
+		pKeyboard->Type = KEYBOARD_TYPE_GUI;
+		gok_keyboard_set_name (pKeyboard, _("GUI"));
+		break;
+	case GOK_SPY_SEARCH_ROLE:
+		pKeyboard->Type = KEYBOARD_TYPE_TOOLBAR; /* catch-all ok ? */
+		/* TODO: localize the return value from AccessibleRole_getName */
+		gok_keyboard_set_name (pKeyboard, AccessibleRole_getName (role));
+		break;
+	case GOK_SPY_SEARCH_APPLICATIONS:
+		pKeyboard->Type = KEYBOARD_TYPE_APPLICATIONS;
+		gok_keyboard_set_name (pKeyboard, _("Applications"));   
+		break;
+	case GOK_SPY_SEARCH_MENU:
+	default:
+		switch (role) {
 		case SPI_ROLE_MENU:
 			pKeyboard->Type = KEYBOARD_TYPE_MENUS;
 			gok_keyboard_set_name (pKeyboard, _("Menus"));
 			break;
 		case SPI_ROLE_MENU_ITEM:
+		default:
 			pKeyboard->Type = KEYBOARD_TYPE_MENUITEMS;
 			gok_keyboard_set_name (pKeyboard, _("Menu"));   
+			pKeyboard->search_type = GOK_SPY_SEARCH_MENU;
 			break;
-		case GOKSPY_ROLE_APPLICATION:
-			pKeyboard->Type = KEYBOARD_TYPE_APPLICATIONS;
-			gok_keyboard_set_name (pKeyboard, _("Applications"));   
-		default:
-			pKeyboard->Type = KEYBOARD_TYPE_GUI;
-			gok_keyboard_set_name (pKeyboard, _("GUI"));
-			break;
+		}
+		break;
 	}
 	
-	pKeyboard->Role = role;
-
 	/* set this flag so the keyboard will be laid out when it's displayed */
 	pKeyboard->bLaidOut = FALSE;
 	pKeyboard->bFontCalculated = FALSE;
 
 	/* display and scan the dynamic keyboard */
 	/* note: keys are added in gok_keyboard_update_dynamic which is */
-	/* called by gok_main_display_scan */
-	gok_main_display_scan ( pKeyboard, pKeyboard->Name, KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL );
+	/* called by gok_main_display_scan */	
+	gok_main_display_scan ( pKeyboard, pKeyboard->Name, 
+		KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL );
 	
 	gok_log_leave();
 	return TRUE;
@@ -1860,11 +1951,14 @@ gboolean gok_keyboard_branch_gui_actions (AccessibleNode* pNodeAccessible)
 	gok_log("checking for children with interesting roles");
 	for (i=0; i< NUM_INTERESTING_ROLES; i++)
 	{
-		if (gok_spy_has_child (pNodeAccessible->paccessible, m_InterestingRole[i]) == TRUE)
+		if (gok_spy_has_child (pNodeAccessible->paccessible, 
+				       GOK_SPY_SEARCH_ROLE, m_InterestingRole[i]) == TRUE)
 		{
 			gok_log_x("Abort action. Branching to child(ren)!");
 			gok_log_leave();
-			return gok_keyboard_branch_gui(pNodeAccessible, m_InterestingRole[i]);
+			return gok_keyboard_branch_gui(pNodeAccessible, 
+						       GOK_SPY_SEARCH_ROLE,
+						       m_InterestingRole[i]);
 		}
 	}
 	gok_log("no children with interesting roles");
@@ -2210,6 +2304,194 @@ void gok_keyboard_insert_array (GokKey* pKey)
 	}
 }
 
+static gboolean
+gok_keyboard_focus_object (Accessible *accessible)
+{
+	gboolean retval = FALSE;
+	AccessibleComponent *component;
+	if (accessible) {
+		component = Accessible_getComponent (accessible);
+		if (component) {
+			retval = AccessibleComponent_grabFocus (component);
+			AccessibleComponent_unref (component);
+		}
+	}
+	return retval;
+}
+
+/* FIXME: this method is a hack necessitated by the fact that link objects aren't
+ * actionable" as they should be, i.e. they don't implement AccessibleAction. 
+ * When that's fixed, we should be able to do away with this kludge
+ * (which manually focussed the text object, places the cursor, and synthesizes
+ * a spacebar key!)
+ */
+static gboolean
+gok_keyboard_follow_link (GokKey *key) 
+{
+	AccessibleHypertext *hypertext = NULL; 
+	AccessibleText *text = NULL;
+	AccessibleNode *node = key->pNodeAccessible;
+	gboolean is_editable = FALSE;
+	if (node && node->paccessible) {
+		hypertext = Accessible_getHypertext (node->paccessible);
+		text = Accessible_getText (node->paccessible);
+		is_editable = Accessible_isEditableText (node->paccessible);
+		/* can't activate editable text this way, it will insert a space instead! */
+		if (hypertext && text && node->is_link && !is_editable) {
+			AccessibleHyperlink *link = AccessibleHypertext_getLink (hypertext, node->link);
+			long int start, end;
+			gboolean focussed;
+			AccessibleHyperlink_getIndexRange (link, &start, &end);
+			focussed = gok_keyboard_focus_object (node->paccessible);
+			if (focussed) {
+				AccessibleText_setCaretOffset (text, start);
+				SPI_generateKeyboardEvent ((long) XK_space, NULL, SPI_KEY_SYM);
+			}
+			AccessibleText_unref (text);
+			AccessibleHypertext_unref (text);
+			return focussed;
+		}
+	}
+	return FALSE;
+}
+
+
+/* TODO: merge with duplicate code in gok-settings-dialog.c */
+
+/**
+ * gok_keyboard_help:
+ *
+ * @pKey: pointer to the invoking GokKey structure. 
+ *
+ * Displays the GOK Help text in the gnome-help browser.
+ *
+ **/
+void
+gok_keyboard_help (GokKey *pKey) 
+{
+	GError *error = NULL;
+	/* TODO: detect error launching help, and give informative message */
+	gnome_help_display ("gok.xml", NULL, &error);
+}
+
+/**
+ * gok_keyboard_about:
+ *
+ * @pKey: pointer to the invoking GokKey structure. 
+ *
+ * Displays the GOK About window.
+ *
+ **/
+void
+gok_keyboard_about (GokKey *pKey) 
+{
+	static GtkWidget *about = NULL;
+	GdkPixbuf	 *pixbuf = NULL;
+	GError		 *error = NULL;
+	gchar		 *file = NULL;
+        
+        static const gchar *authors [] = {
+		"Simon Bates  <simon.bates@utoronto.ca>",
+		"David Bolter <david.bolter@utoronto.ca>",
+		"Bill Haneman <bill.haneman@sun.com>",
+		"Chris Ridpath <chris.ridpath@utoronto.ca>",
+		NULL
+	};
+
+	const gchar *documenters[] = {
+		NULL
+	};
+
+	const gchar *translator_credits = _("translator_credits");
+
+	if (about) {
+		gtk_window_set_screen (
+			GTK_WINDOW (about),
+			gtk_widget_get_screen (GTK_WIDGET (gok_main_get_main_window ())));
+		gtk_window_present (GTK_WINDOW (about));
+		return;
+	}
+        
+	file = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_APP_DATADIR, 
+					  "gok/goklogo.png", 
+					  FALSE, NULL);
+	if (file) {
+		pixbuf = gdk_pixbuf_new_from_file (file, &error);
+		g_free (file);
+	}
+
+	if (error) {
+		g_warning (G_STRLOC ": cannot open %s: %s", file, error->message);
+		g_error_free (error);
+	}
+	
+        about = gnome_about_new (
+		_("GOK"), VERSION,
+		_("Copyright (C) 2001-2003 Sun Microsystems, Copyright (C) 2001-2003 University of Toronto"),
+		_("Dynamic virtual keyboards for the GNOME desktop"),
+		authors,
+		documenters,
+		strcmp (translator_credits, "translator_credits") != 0 ? translator_credits : NULL,
+		pixbuf);
+		
+	if (pixbuf)
+		gdk_pixbuf_unref (pixbuf);
+			
+	gtk_window_set_wmclass (GTK_WINDOW (about), "gok", "GOK");
+	gtk_window_set_screen (GTK_WINDOW (about),
+			       gtk_widget_get_screen (GTK_WIDGET (gok_main_get_main_window ())));
+	g_signal_connect (about, "destroy",
+			  G_CALLBACK (gtk_widget_destroyed),
+			  &about);
+        gtk_widget_show (about);
+}
+
+/**
+ * gok_keyboard_dock:
+ *
+ * @pKey: pointer to the GokKey structure whose activation
+ *        should relocate the GokKeyboard's docking position 
+ *        onscreen.
+ *
+ * Docks the GOK keyboard to the top or bottom of the screen,
+ *        or "floats" the keyboard if "none" is specified as the
+ *        docking direction.
+ *
+ **/
+void
+gok_keyboard_dock (GokKey *pKey) 
+{
+	GokKeyboardDirection dir = GOK_DIRECTION_NONE;
+	GokDockType old_type = gok_data_get_dock_type ();
+	gint top = 0, bottom = 0;
+
+	if (pKey && pKey->pGeneral) 
+		dir = *(GokKeyboardDirection *) pKey->pGeneral;
+	switch (dir)
+	{
+	case GOK_DIRECTION_N:
+		gok_data_set_dock_type (GOK_DOCK_TOP);
+		gok_main_set_wm_dock (TRUE);
+		break;
+	case GOK_DIRECTION_S:
+		gok_data_set_dock_type (GOK_DOCK_BOTTOM);
+		gok_main_set_wm_dock (TRUE);
+		break;
+	default:
+		gok_main_update_struts (0, 0);
+		gok_data_set_dock_type (GOK_DOCK_NONE);
+		gok_main_set_wm_dock (FALSE);
+	        break;
+	}
+	if (dir == GOK_DIRECTION_NONE) {
+		GtkWidget *widget = gok_main_get_main_window ();
+		if (widget) 
+			gdk_window_raise (widget->window); 
+/* FIXME: why do we need this? */ 
+		gtk_widget_show_now (gok_main_get_main_window ());
+	}
+}
+
 /**
  * gok_keyboard_move_resize:
  *
@@ -2253,6 +2535,12 @@ gok_keyboard_move_resize (GokKey *pKey)
 	case GOK_DIRECTION_S:
 		y += 10;
 		break;
+	case GOK_DIRECTION_FILL_EW:
+		/* toggle expand */
+		fprintf (stderr, "toggling fill\n");
+		gok_data_set_expand (!gok_data_get_expand ());
+		gok_main_display_scan_reset ();
+		break;
 	default:
 	        break;
 	}
@@ -2318,11 +2606,18 @@ int gok_keyboard_output_selectedkey (void)
 		case KEYTYPE_BRANCHTOOLBARS:
 		case KEYTYPE_BRANCHGUI:
 		case KEYTYPE_BRANCHEDITTEXT:
+	        case KEYTYPE_BRANCHHYPERTEXT:
 			gok_keyboard_branch_byKey (pKeySelected);
 			break;
 
 		case KEYTYPE_BRANCHALPHABET:
-			gok_main_display_scan ( NULL, "alphabet", KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL);
+			gok_main_display_scan ( NULL, CORE_KEYBOARD, KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL);
+			break;
+
+		case KEYTYPE_BRANCHTEXT:
+			if (pKeySelected->pNodeAccessible && pKeySelected->pNodeAccessible->paccessible)
+				gok_keyboard_focus_object (pKeySelected->pNodeAccessible->paccessible);
+			gok_main_display_scan ( NULL, "Keyboard", KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL);
 			break;
 
 		case KEYTYPE_SETTINGS:
@@ -2342,6 +2637,24 @@ int gok_keyboard_output_selectedkey (void)
 	        case KEYTYPE_MOVERESIZE:	
 		        gok_keyboard_move_resize (pKeySelected);
 		        break;
+
+	        case KEYTYPE_DOCK:	
+		        gok_keyboard_dock (pKeySelected);
+		        break;
+
+	        case KEYTYPE_HELP:	
+		        gok_keyboard_help (pKeySelected);
+		        break;
+
+	        case KEYTYPE_ABOUT:	
+		        gok_keyboard_about (pKeySelected);
+		        break;
+
+	        case KEYTYPE_HYPERLINK:
+		        if (gok_keyboard_follow_link (pKeySelected))
+				gok_main_display_scan_reset ();
+			break;
+
 		default:
 			/* no branching, just a regular key */
 			gok_keyboard_output(pKeySelected);
@@ -2400,7 +2713,8 @@ void gok_keyboard_output(GokKey* pKey)
 	case KEYTYPE_BRANCHMENUITEMS:
 	case KEYTYPE_MENUITEM:
 	case KEYTYPE_BRANCHGUIACTIONS:
-		gok_keyboard_branch_byKey (pKey);
+		if (!gok_keyboard_branch_byKey (pKey))
+			gok_main_display_scan_reset ();
 		break;
 		
 	case KEYTYPE_EDITABLETEXT_CUT:
@@ -2415,6 +2729,7 @@ void gok_keyboard_output(GokKey* pKey)
 		gok_modifier_output_pre();
 		
 		/* TODO: add check to see if command prediciton turned on ?*/
+#ifdef WHENCOMMANDPREDICTIONINGOKDATAISHOOKEDUPTOPREFS
 		{			
 			pPredictedKeyboard = NULL;
 			pKeyboardTemp = gok_main_get_first_keyboard();
@@ -2442,7 +2757,7 @@ void gok_keyboard_output(GokKey* pKey)
 				gok_predictor_log(gok_main_get_command_predictor(),gok_keyboard_get_name(pPredictedKeyboard), gok_key_get_label(pKey) );
 			}
 		}
-		
+#endif /* WHENCOMMANDPREDICTIONINGOKDATAISHOOKEDUPTOPREFS */
 		/* send the key output to the system */
 		gok_output_send_to_system (pKey->pOutput, TRUE);
 		
@@ -2529,7 +2844,8 @@ gboolean gok_keyboard_validate_dynamic_keys (Accessible* pAccessibleForeground)
 	{
 		if (pKey->Type == KEYTYPE_BRANCHMENUS)
 		{
-			if (gok_spy_has_child (pAccessibleForeground, SPI_ROLE_MENU) == TRUE)
+			if (gok_spy_has_child (pAccessibleForeground, 
+					       GOK_SPY_SEARCH_ROLE, SPI_ROLE_MENU) == TRUE)
 			{
 				/* set this flag if the button state changes */
 				if (strcmp (gtk_widget_get_name(pKey->pButton), "StyleButtonBranchMenus") != 0)
@@ -2556,7 +2872,8 @@ gboolean gok_keyboard_validate_dynamic_keys (Accessible* pAccessibleForeground)
 		}
 		else if (pKey->Type == KEYTYPE_BRANCHTOOLBARS)
 		{
-			if (gok_spy_has_child (pAccessibleForeground, SPI_ROLE_TOOL_BAR) == TRUE)
+			if (gok_spy_has_child (pAccessibleForeground, GOK_SPY_SEARCH_ROLE,
+					       SPI_ROLE_TOOL_BAR) == TRUE)
 			{
 				/* set this flag if the button state changes */
 				if (strcmp (gtk_widget_get_name(pKey->pButton), "StyleButtonBranchToolbars") != 0)
@@ -2583,7 +2900,11 @@ gboolean gok_keyboard_validate_dynamic_keys (Accessible* pAccessibleForeground)
 		}
 		else if (pKey->Type == KEYTYPE_BRANCHGUI)
 		{
-			if (gok_spy_has_child (pAccessibleForeground, SPI_ROLE_PUSH_BUTTON) == TRUE)
+			/* TODO: REFACTOR: enum KeyType */
+			if ((gok_spy_has_child (pAccessibleForeground, 
+						GOK_SPY_SEARCH_ROLE, SPI_ROLE_PUSH_BUTTON) == TRUE) ||
+				(gok_spy_has_child (pAccessibleForeground, 
+						    GOK_SPY_SEARCH_EDITABLE_TEXT, SPI_ROLE_TEXT) == TRUE))
 			{
 				/* set this flag if the button state changes */
 				if (strcmp (gtk_widget_get_name(pKey->pButton), "StyleButtonBranchGUI") != 0)
@@ -2707,7 +3028,8 @@ void gok_keyboard_on_window_resize ()
 	{
 		widthKey = MAX_KEY_WIDTH;
 	}
-	gok_data_set_key_width (widthKey);
+	if ((gok_data_get_dock_type () == GOK_DOCK_NONE) || (pKeyboard->expand == GOK_EXPAND_NEVER)) 
+		gok_data_set_key_width (widthKey);
 
 	heightKey = gok_keyboard_get_keyheight_for_window (heightWindow, pKeyboard);
 	if (heightKey < MIN_KEY_HEIGHT)
@@ -2718,6 +3040,7 @@ void gok_keyboard_on_window_resize ()
 	{
 		heightKey = MAX_KEY_HEIGHT;
 	}
+	if ((gok_data_get_dock_type () == GOK_DOCK_NONE) || (pKeyboard->expand == GOK_EXPAND_NEVER)) 
 	gok_data_set_key_height (heightKey);
 	
 	/* set this flag on each keyboard so it will calculate a new font size */
@@ -2859,5 +3182,7 @@ gok_keyboard_parse_direction (const gchar *string)
 		return GOK_DIRECTION_S;
 	else if (!strcmp (string, "southeast"))
 		return GOK_DIRECTION_SE;
+	else if (!strcmp (string, "fillwidth"))
+		return GOK_DIRECTION_FILL_EW;
 	return GOK_DIRECTION_NONE;
 }
