@@ -78,12 +78,15 @@ SPIBoolean gok_spy_key_listener (const AccessibleKeystroke *key,
                                  void *user_data);
 
 AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
+				     GokSpySearchType type,
                                      AccessibleRole anAccessiblerole);
 
-static void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole Role);
+static void gok_spy_find_children (AccessibleNode* pNode, GokSpySearchType type, AccessibleRole Role);
 
 static gboolean gok_spy_has_child_aux (Accessible* accessible,
-                                       AccessibleRole role, gint children,
+				       GokSpySearchType search_type,
+                                       AccessibleRole role,
+				       gint children,
                                        gint depth, gint cdepth,
                                        gboolean* found);
 
@@ -221,11 +224,10 @@ gboolean gok_spy_accessible_is_okay(Accessible* accessible)
 {
 	gboolean returncode = FALSE;
 	gok_log_enter();
-	if (accessible != NULL)
+	if (accessible != NULL) 
 	{
-		/*
+		/* FIXME: we should have a better way of doing this! */
 		returncode = cspi_accessible_is_a (accessible, "IDL:Accessibility/Accessible:1.0");
-		*/
 		/*
 		if (Accessible_isApplication(accessible) ||
 		    Accessible_isComponent(accessible))
@@ -243,11 +245,6 @@ gboolean gok_spy_accessible_is_okay(Accessible* accessible)
 				returncode = TRUE;
 			}
 			*/
-		/* kludge - check to see if there is even a window */
-		if (m_ptheWindowAccessible != NULL)
-		{
-			returncode = TRUE;
-		}
 	}
 	gok_log("returning: %d",returncode);
 	gok_log_leave();
@@ -311,6 +308,10 @@ void gok_spy_open()
 			windowDeactivateListener, "window:deactivate");
 		gok_log ("SPI_registerGlobalEventListener for window:deactivate returned: %d", success);
 
+		success = SPI_registerGlobalEventListener (
+			windowDeactivateListener, "window:destroy");
+		gok_log ("SPI_registerGlobalEventListener for window:destroy returned: %d", success);
+
 	}
 	gok_log_leave ();
 }
@@ -450,23 +451,22 @@ void gok_spy_deregister_mousebuttonlistener(MouseButtonListener* callback)
 * gok_spy_get_list 
 * 
 * @paccessible: The parent accessible to the list
+* @type: The type of list to return, i.e. the search criteria
 * @role: The role of the list members
 *
 * User must call gok_spy_free when finished with this list.
 * 
 * Returns: pointer to the list or NULL
 **/ 
-AccessibleNode* gok_spy_get_list( Accessible* paccessible, AccessibleRole role )
+AccessibleNode* gok_spy_get_list( Accessible* paccessible, GokSpySearchType type, AccessibleRole role )
 {
 	AccessibleNode* plist = NULL;
 	gok_log_enter();
-	plist = gok_spy_create_list( paccessible, role );
+	plist = gok_spy_create_list( paccessible, type, role );
 	gok_log_leave();
 	return plist;
 }
 
-
-
 /** 
 * gok_spy_refresh 
 * 
@@ -504,8 +504,8 @@ void gok_spy_free (AccessibleNode* pNode)
 	{
 		pNodeToFree = pNode;
 		pNode = pNode->pnext;
-		
-		SPI_freeString (pNodeToFree->pname);
+		/* we gfree this since we've dup'd the SPI string */
+		g_free (pNodeToFree->pname);
 		gok_spy_accessible_unref (pNodeToFree->paccessible);
 		g_free (pNodeToFree);
 	}
@@ -532,13 +532,14 @@ Accessible* gok_spy_get_accessibleWithEditableText()
 * 
 * Returns: true if a child exists, false otherwise. 
 **/ 
-gboolean gok_spy_has_child (Accessible* accessible, AccessibleRole role) 
+gboolean gok_spy_has_child (Accessible* accessible, GokSpySearchType type, AccessibleRole role) 
 {
 	gboolean bfound = FALSE;
 	gok_log_enter();
 	if (gok_spy_accessible_is_okay(accessible) == TRUE)
 	{
-		gok_spy_has_child_aux (accessible, role, GOKSPY_GUI_BREADTH, GOKSPY_GUI_DEPTH, 0, &bfound);
+		gok_spy_has_child_aux (accessible, type, role, 
+				       GOKSPY_GUI_BREADTH, GOKSPY_GUI_DEPTH, 0, &bfound);
 	}
 	else
 	{
@@ -588,6 +589,7 @@ void gok_spy_check_queues(void)
 			break;
 		case GOKSPY_WINDOW_ACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_ACTIVATE_EVENT");
+			/* FIXME !! the ref needs to happen in the original event loop */
 			if (en->paccessible != m_ptheWindowAccessible) {
 				gok_spy_accessible_unref (m_ptheWindowAccessible);
 				m_ptheWindowAccessible = en->paccessible;
@@ -607,11 +609,7 @@ void gok_spy_check_queues(void)
 		case GOKSPY_WINDOW_DEACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_DEACTIVATE_EVENT");
 
-			/* TODO: Determine if this workaround is necessary
-			 *       or if at-spi should be modified to
-			 *       not send out of order activate/deactivate
-			 *       events
-			 *
+			/*
 			 * The test below is needed to workaround the case
 			 * that a window:deactivate event may arrive
 			 * after a window:activate has been emitted by
@@ -675,10 +673,11 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
 	ass = Accessible_getStateSet(accessible);
 	if (ass != NULL)
 	{
+		AccessibleRole role = Accessible_getRole (accessible);
 		/* assumption: if this thing manages its own children, then it 
 		 * is probably not worth traversing.
 		 */
-		if ( AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ) == TRUE)
+		if ( AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ))
 		{
 			worthiness = FALSE;
 		}
@@ -692,7 +691,8 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
 /** 
  * gok_spy_has_child_aux:
  * @accessible: Pointer to the accessible.
- * @role: The role you are searching for.
+ * @type: The search criteria to apply.
+ * @role: The role you are searching for, if @type is GOK_SPY_SEARCH_ROLE.
  * @children: the breadth of the gui tree to search (for optimization).
  * @depth: how deep to look in the gui tree (for optimization).
  * @cdepth: recursive variable that keeps track of current depth
@@ -704,10 +704,14 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
  *
  * Returns: true if a child exists, false otherwise.
  */
-gboolean gok_spy_has_child_aux (Accessible* accessible, AccessibleRole role, gint children, gint depth, gint cdepth, gboolean* found) 
+gboolean gok_spy_has_child_aux (Accessible* accessible, 
+				GokSpySearchType type,
+				AccessibleRole role, 
+				gint children, gint depth, gint cdepth, gboolean* found) 
 { 
     Accessible* pAccessibleChild; 
-	int child_count;
+    AccessibleStateSet *states;
+    int child_count;
     int i; 
 
     if (accessible == NULL) return *found; 
@@ -733,29 +737,71 @@ gboolean gok_spy_has_child_aux (Accessible* accessible, AccessibleRole role, gin
 		if (pAccessibleChild != NULL) 
 		{ 
 			/* does this child have the role we're looking for? */ 
-		    if (Accessible_getRole (pAccessibleChild) == role) 
+		    if (type == GOK_SPY_SEARCH_ROLE) 
 		    {
-		    	gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
-		    	gok_spy_accessible_unref(pAccessibleChild);
-		    	*found = TRUE;
-	    		return *found;
-		    } 
-		    else /* search through the children of this child */ 
-		    { 
-		    	if (*found == TRUE)
-		    	{
-			    	gok_spy_accessible_unref(pAccessibleChild);
-		    		return *found;
-		    	}
-		    	gok_spy_has_child_aux (pAccessibleChild, role, children, depth, cdepth, found);
-		    	gok_spy_accessible_unref(pAccessibleChild);
-		    } 
+			    if (Accessible_getRole (pAccessibleChild) == role) {
+				    gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
+				    gok_spy_accessible_unref(pAccessibleChild);
+				    *found = TRUE;
+				    return *found;
+			    }
+		    }
+		    else if (type == GOK_SPY_SEARCH_EDITABLE_TEXT) {
+			    if (Accessible_isEditableText(pAccessibleChild)) {
+				    *found = TRUE;
+				    gok_spy_accessible_unref(pAccessibleChild);
+				    return *found;
+			    }
+		    }
+		    gok_spy_has_child_aux (pAccessibleChild, type, role, children, depth, cdepth, found);
+		    gok_spy_accessible_unref(pAccessibleChild);
 		} 
 	}
 	
 	return *found;
 } 
 
+static gboolean
+gok_spy_is_visible (Accessible *accessible)
+{
+	gboolean retval = FALSE;
+	if (accessible) {
+		AccessibleStateSet *states = Accessible_getStateSet (accessible);
+		if (AccessibleStateSet_contains (states, SPI_STATE_VISIBLE) &&
+		    AccessibleStateSet_contains (states, SPI_STATE_SHOWING))
+			retval = TRUE;
+		AccessibleStateSet_unref (states);
+	}
+	return retval;
+}
+
+
+static gboolean
+gok_spy_is_ui (Accessible *accessible, AccessibleRole role)
+{
+	gboolean is_interesting_ui_component = FALSE;
+	switch (role) {
+	case SPI_ROLE_PUSH_BUTTON:
+	case SPI_ROLE_CHECK_BOX:
+	case SPI_ROLE_COMBO_BOX:
+	case SPI_ROLE_RADIO_BUTTON:
+	case SPI_ROLE_TOGGLE_BUTTON:
+		if (gok_spy_is_visible (accessible))
+			is_interesting_ui_component = TRUE;
+		
+	default:
+		break;
+	}
+	/* EditableText and Hypertext components are always interesting */
+	if (!is_interesting_ui_component) {
+		if ((Accessible_isEditableText (accessible)) ||
+		    (Accessible_isHypertext (accessible)))
+			if (gok_spy_is_visible (accessible)) 
+				is_interesting_ui_component = TRUE;
+	}
+	return is_interesting_ui_component;
+}
+
 /**
  * gok_spy_find_children:
  * @pNode: Before execution of this function it is the node that you
@@ -770,7 +816,7 @@ gboolean gok_spy_has_child_aux (Accessible* accessible, AccessibleRole role, gin
  * beginning at pNode->pnext.  Please note that this function may also
  * modify pNode->paccessible.  This is a recursive function.
  */
-void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole role) 
+void gok_spy_find_children (AccessibleNode* pNode, GokSpySearchType type, AccessibleRole role) 
 { 
 	Accessible* pAccessibleChild;
 	AccessibleRole childrole;
@@ -801,24 +847,30 @@ void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole role)
 
 		if (pAccessibleChild != NULL)
 		{
-			/* does this child have the role we're looking for? */
 			childrole = Accessible_getRole (pAccessibleChild);
 
-			/* special case menu stuff */
-			if (((role == SPI_ROLE_MENU_ITEM) && 
-			    ((childrole ==  SPI_ROLE_MENU_ITEM) ||
-			     (childrole ==  SPI_ROLE_CHECK_MENU_ITEM) ||
-			     (childrole ==  SPI_ROLE_RADIO_MENU_ITEM) ||
-			     (childrole ==  SPI_ROLE_MENU))) 
-				||
-				(childrole == role))
+			/* check role: special-casing some stuff REFACTOR THIS later*/
+			if ((type == GOK_SPY_SEARCH_ALL) ||
+			    ((type == GOK_SPY_SEARCH_MENU) && 
+			     ((childrole ==  SPI_ROLE_MENU_ITEM) ||
+			      (childrole ==  SPI_ROLE_CHECK_MENU_ITEM) ||
+			      (childrole ==  SPI_ROLE_RADIO_MENU_ITEM) ||
+			      (childrole ==  SPI_ROLE_MENU))) 
+			    ||
+			    ((type == GOK_SPY_SEARCH_ROLE) && (childrole == role)) 
+			    ||
+			    ((type == GOK_SPY_SEARCH_UI) && 
+			     (gok_spy_is_ui (pAccessibleChild, childrole)))
+			    ||
+			    ((type == GOK_SPY_SEARCH_EDITABLE_TEXT) &&
+			     (Accessible_isEditableText(pAccessibleChild))))
 			{
 				gok_spy_append_node (pNode, pAccessibleChild);
 			}
 			else /* search through the children of this child */
 			{
 				pNode->paccessible = pAccessibleChild;
-				gok_spy_find_children (pNode, role);
+				gok_spy_find_children (pNode, type, role);
 			}
 
 		gok_spy_accessible_unref (pAccessibleChild);
@@ -838,6 +890,7 @@ void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole role)
  * children found.
  */
 AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
+				     GokSpySearchType type,
                                      AccessibleRole anAccessibleRole)
 {
 	AccessibleNode nodeRoot;
@@ -850,12 +903,35 @@ AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
 	nodeRoot.pnext = NULL;
 	nodeRoot.pname = NULL;
 
-	gok_spy_find_children (&nodeRoot, anAccessibleRole);
+	gok_spy_find_children (&nodeRoot, type, anAccessibleRole);
 
 	gok_log_leave();
 	return nodeRoot.pnext;
 }
 
+static gboolean
+gok_spy_insert_node (AccessibleNode *prev, Accessible *accessible, gint i, char *name)
+{
+	AccessibleNode *newnode = NULL;
+
+	if (!prev) return FALSE;
+
+	newnode = (AccessibleNode*)g_malloc(sizeof(AccessibleNode));
+	
+	if (newnode != NULL) {
+		newnode->paccessible = accessible;
+		gok_spy_accessible_ref (accessible);
+		newnode->is_link = (i >= 0) ? TRUE : FALSE;
+		newnode->link = MAX (0, i);
+		newnode->pname = name;
+		gok_log("newnode->pname = %s", newnode->pname);
+		newnode->pnext = prev->pnext;	
+		prev->pnext = newnode;
+		return TRUE;
+	}
+	return FALSE;
+}
+
 /**
  * gok_spy_append_node:
  * @pNode: An existing list of AccessibleNode to append to, must be
@@ -876,49 +952,153 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 {
 	AccessibleNode* pLastNode = NULL;
 	AccessibleNode* pNewNode = NULL;
-	char* pName = NULL;
+	AccessibleRelation** relations = NULL;
+	Accessible* targetAccessible = NULL;
+	AccessibleRelationType type = -1;
+	char* pName = NULL, *s;
+	int i = 0;
+	int j = 0;
+	int ntargets = 0;
 
 	g_assert (pNode != NULL);
 
 	gok_log_enter();
 
-	pName = Accessible_getName (pAccessible);
+        s = Accessible_getName (pAccessible);
+	if (s && strlen (s)) 
+	  pName = g_strdup (s);
+	if (s)
+	  SPI_freeString (s);
 
-	/* if the name is empty then ignore this child */ 
-	if ( (pName !=NULL) && (strlen (pName) != 0) )
-	{
-		/* create a new node */
-		pNewNode = (AccessibleNode*)g_malloc(sizeof(AccessibleNode));
-
-		if (pNewNode != NULL)
-		{
-			pNewNode->paccessible = pAccessible;
-			gok_spy_accessible_ref (pAccessible);
-
-			pNewNode->pname = pName;
-			gok_log("pNewNode->pname = %s",pNewNode->pname);
-			pNewNode->pnext = NULL;
-
-			pLastNode = pNode;
-			while (pLastNode->pnext != NULL)
-			{
-				pLastNode = pLastNode->pnext;
+	/* Hypertext objects are treated differently; one node per link */
+	if (Accessible_isHypertext (pAccessible)) {
+		AccessibleHypertext *hypertext = Accessible_getHypertext (pAccessible);
+		AccessibleText *text = Accessible_getText (pAccessible);
+		gint i, nlinks = AccessibleHypertext_getNLinks (hypertext);
+		for (i = 0; i < nlinks; ++i) {
+			AccessibleHyperlink *link = AccessibleHypertext_getLink (hypertext, i);
+			Accessible *link_object = AccessibleHyperlink_getObject (link, 0);
+			gchar *linkname;
+			long int start, end;
+			/* TODO: handle multi-anchor links! */
+			AccessibleHyperlink_getIndexRange (link, &start, &end);
+			linkname = AccessibleText_getText (text, start, end);
+			if (linkname) {
+				/* create and append a node */
+			  /* 
+			   * FIXME: when AccessibleHyperlink objects get un-broken,
+			   * we should pass the hyperlink object here instead of the 
+			   * hypertext object, so we can just invoke the default action
+			   * on the link!  But at the moment the link objects are not
+			   * actionable, so we have to fake it.
+			   */
+				gok_spy_insert_node (pNode, pAccessible, i, linkname);
 			}
-			pLastNode->pnext = pNewNode;
+			Accessible_unref (link_object);
 		}
-		else
+		AccessibleText_unref (text);
+		AccessibleHypertext_unref (hypertext);
+	}
+
+	/* if the name is empty then try LABELLED_BY, else ignore this child */ 
+	else {
+		/* look for a label */
+		if (pName == NULL)
 		{
-			SPI_freeString (pName);
+			gok_log("no name, so looking at relations...");
+			relations = Accessible_getRelationSet(pAccessible);
+			if (relations == NULL)	{
+				gok_log("no relationset for this object");
+				gok_log_leave();
+				return pNewNode;
+			}
+			while (relations[i]) {
+				type = AccessibleRelation_getRelationType(relations[i]);
+				if (type == SPI_RELATION_LABELED_BY) {
+					ntargets = AccessibleRelation_getNTargets(relations[i]);
+					for (j=0; j < ntargets; j++) {
+						targetAccessible = 
+							AccessibleRelation_getTarget(relations[i], j);
+						if (targetAccessible != NULL) {
+							s = Accessible_getName (targetAccessible);
+							
+							if (s != NULL) {
+								if ( strlen (s) != 0 )  {
+									pName = g_strdup (s);
+									SPI_freeString (s);
+									break; /* from for loop*/
+								}
+								else {
+									SPI_freeString(s);
+								}
+							}
+						}
+					}
+					break; /* from while loop */
+				}
+			}
+		}
+		
+		if (pName != NULL) {
+			/* create a new node */
+			gok_spy_insert_node (pNode, pAccessible, -1, pName);
+		}
+		else if (Accessible_isText (pAccessible)) {
+			/* 
+			 * sadly, many text objects are still nameless - yet we want 
+			 * to allow the user to get to them.
+			 * first tack is to get the starting text... if the object is
+			 * empty, then we can still "create" a name for it, though we 
+			 * can't assign it a unique meaningful name.
+			 */
+			AccessibleText *text = Accessible_getText (pAccessible);
+			gint j = 0, len;
+			gchar *word, *gs = NULL;
+		        if (text) {
+				long int start, end = 0;
+				do {
+					gchar *sp = gs;
+					word = AccessibleText_getTextAtOffset (text, end, 
+									       SPI_TEXT_BOUNDARY_WORD_END,
+									       &start, &end);
+					if (word) 
+						len = strlen (word);
+					else 
+						len = 0;
+					if (len) {
+						if (sp) {
+							gs = g_strconcat (sp, word, NULL);
+							g_free (sp);
+						}
+						else
+							gs = g_strdup (word);
+					}
+					if (word) 
+						SPI_freeString (word);
+					++j;
+				} while ((j < 3) && len);
+				AccessibleText_unref (text);
+			}
+			if (gs && strlen (gs)) 
+				pName = g_strconcat (gs, "...", NULL);
+			else {
+			        if (gs) 
+					g_free (gs);
+				if (Accessible_isEditableText (pAccessible))
+					pName = g_strdup ("Text Entry (empty)");
+				else
+					pName = g_strdup ("Text (empty)");
+			}
+			gok_spy_insert_node (pNode, pAccessible, -1, pName);
 		}
 	}
-	else if (pName != NULL)
-	{
-		SPI_freeString (pName);
-	}
+	
+	while (pNode->pnext) 
+		pNode = pNode->pnext;
 
 	gok_log_leave();
 
-	return pNewNode;
+	return pNode;
 }
 
 /** 
