@@ -34,6 +34,16 @@ extern "C" {
 
 /*G_LOCK_DEFINE_STATIC(gok_spy_mutex);*/
 #define GOKSPY_ROLE_APPLICATION -1
+#define GOKSPY_ROLE_EDITABLETEXT -2
+
+typedef enum {
+	GOK_SPY_SEARCH_ROLE,
+	GOK_SPY_SEARCH_MENU,
+	GOK_SPY_SEARCH_UI,
+	GOK_SPY_SEARCH_EDITABLE_TEXT,
+	GOK_SPY_SEARCH_APPLICATIONS,
+	GOK_SPY_SEARCH_ALL
+} GokSpySearchType;
 
 void gok_spy_open(void);        /* call this first (user must call SPI_init first) */
 void gok_spy_close(void);       /* call this when shutting down */ 
@@ -53,13 +63,14 @@ gulong gok_spy_get_modmask (void);
 /* gboolean gok_spy_is_desktop(Accessible* pAccessible); */
 /* gboolean gok_spy_accessible_is_desktopChild(Accessible* accessible);*/
 /* AccessibleNode* gok_spy_get_applications( void ); */
-AccessibleNode* gok_spy_get_list( Accessible* paccessible, AccessibleRole role ); 
+AccessibleNode* gok_spy_get_list( Accessible* paccessible, GokSpySearchType type, AccessibleRole role ); 
+AccessibleNode* gok_spy_get_ui_list( Accessible* paccessible); 
 AccessibleNode* gok_spy_refresh( AccessibleNode* plist); 
 void gok_spy_free( AccessibleNode* plist ); /* call this to free up an uneeded list's memory */ 
 
 Accessible* gok_spy_get_accessibleWithEditableText(void);
 
-gboolean gok_spy_has_child (Accessible* accessible, AccessibleRole role); 
+gboolean gok_spy_has_child (Accessible* accessible, GokSpySearchType type, AccessibleRole role); 
 
 void gok_spy_check_queues(void);
 gboolean gok_spy_check_window(AccessibleRole role);
@@ -76,5 +87,4 @@ void gok_spy_accessible_unref(Accessible* accessible);
 } 
 #endif /* __cplusplus */ 
  
-#endif // #ifndef __GOKSPY_H__ 
- 
+#endif // #ifndef __GOKSPY_H__
