@@ -487,7 +487,8 @@ void gok_main_window_change_listener (Accessible* pAccessible)
 	gok_log_enter();	
 	if (1) /* TODO check to make sure the new window is not the gok window! */
 	{
-		m_pForegroundWindowAccessible = pAccessible;
+		/* note: ref counting happens in gok-spy's listeners */
+	        m_pForegroundWindowAccessible = pAccessible;
 		
 		/* we might currently have a dynamic keyboard so we need to branch to be safe */
 		gok_main_display_scan_previous_premade();
@@ -984,6 +985,10 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 	/* stop the scanning process */
 	gok_scanner_stop();
 
+	/* raise the window as a precaution */
+	if (m_pWindowMain && m_pWindowMain->window)
+		gdk_window_raise (m_pWindowMain->window);
+
 	/* create new keys for dynamic keyboards */
 	if (pKeyboard->bDynamicallyCreated == TRUE)
 	{
@@ -1388,6 +1393,7 @@ void gok_main_initialize_commandprediction ()
 /**
 * gok_main_resize_window
 * @pWindow: Pointer to the main window.
+* @pKeyboard: Pointer to the relevant (pending) keyboard (may be NULL).
 * @Width: Width of the new window.
 * @Height: Height of the new window.
 * 
@@ -1399,7 +1405,7 @@ void gok_main_initialize_commandprediction ()
 * window then m_countIgnoreConfigure will be set to 2 so that the window 
 * center is not changed.
 */
-void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
+void gok_main_resize_window (GtkWidget* pWindow, GokKeyboard *pKeyboard, gint Width, gint Height)
 {
 	gint left;
 	gint top;
@@ -1430,13 +1436,15 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 	
 	/* let the keyboard know that the resize was caused by us */
 	gok_keyboard_set_ignore_resize (TRUE);
+
 	
 	/* calculate the upper left corner of the window */
-	switch (gok_data_get_dock_type ()) { /* if we're a dock */
+	switch (gok_data_get_dock_type () ) { /* if we're a dock */
 	case GOK_DOCK_BOTTOM:
 	  top = gdk_screen_height ();
 	  left = 0;
-	  if (gok_data_get_expand ()) {
+	  if ((pKeyboard->expand == GOK_EXPAND_ALWAYS) || 
+	      (gok_data_get_expand () && (pKeyboard->expand != GOK_EXPAND_NEVER))) {
 		  Width = gdk_screen_width ();
 		  gok_keyboard_set_ignore_resize (FALSE);
 	  }
@@ -1444,7 +1452,8 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 	case GOK_DOCK_TOP:
 	  top = 0;
 	  left = 0;
-	  if (gok_data_get_expand ()) {
+	  if ((pKeyboard->expand == GOK_EXPAND_ALWAYS) || 
+	      (gok_data_get_expand () && (pKeyboard->expand != GOK_EXPAND_NEVER))) {
 		  Width = gdk_screen_width ();
 		  gok_keyboard_set_ignore_resize (FALSE);
 	  }
@@ -1519,14 +1528,47 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 	/* move/resize the main window */
 	gdk_window_move_resize ((GdkWindow*)pWindow->window, left, top, Width, Height);
 
-	if (gok_data_get_dock_type () != GOK_DOCK_NONE)
-		gok_main_update_struts (Width, Height);
+	if (gok_data_get_dock_type () != GOK_DOCK_NONE) {
+		/* if this is a "never expand" keyboard, unset struts! */
+		if (pKeyboard && (pKeyboard->expand == GOK_EXPAND_NEVER))
+			gok_main_update_struts (Width, 1);
+		else
+			gok_main_update_struts (Width, Height);
+	}
 
 	/* store these values */
 	m_OurResizeWidth = Width;
 	m_OurResizeHeight = Height;
 }
 
+void
+gok_main_set_wm_dock (gboolean is_dock)
+{
+  Atom atom_type[1], atom_window_type;
+  GtkWidget *widget = gok_main_get_main_window ();
+
+  gdk_error_trap_push ();
+  atom_window_type = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE");
+  if (is_dock) 
+	  atom_type[0] = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE_DOCK");
+  else
+	  atom_type[0] = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE_NORMAL");
+
+  XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
+		   GDK_WINDOW_XWINDOW (widget->window), 
+		   atom_window_type,
+		   XA_ATOM, 32, PropModeReplace,
+		   (guchar *) &atom_type, 1);
+  gdk_error_trap_pop ();
+
+  if (m_pCurrentKeyboard) {
+	  gok_keyboard_display (m_pCurrentKeyboard, m_pCurrentKeyboard,
+				gok_main_get_main_window(), TRUE); 
+          /* this should move us to appropriate screen edge */
+  }
+
+}
+
 /*
  * TODO: GtkDoc this!
  */
@@ -1537,7 +1579,7 @@ gok_main_update_struts (gint width, gint height)
 	Atom atom_strut = gdk_x11_get_xatom_by_name ("_NET_WM_STRUT");
 	guint32 struts[4];
 
-	switch (gok_data_get_dock_type () != GOK_DOCK_NONE) {
+	switch (gok_data_get_dock_type ()) {
 	case GOK_DOCK_TOP:
 		struts[0] = struts[2] = struts[3] = 0;
 		struts[1] = height;
