@@ -33,7 +33,7 @@ extern "C" {
 #include "gok-keyboard.h"
 
 /* public */
-GtkWidget* gok_main_create_window (void);
+GtkWidget* gok_main_create_window (gboolean is_dock);
 void gok_main_set_cursor (GdkCursor *cursor);
 gint gok_main_open(gint argc, gchar *argv[]);
 gboolean gok_main_display_scan (GokKeyboard* pKeyboard, char* NameKeyboard, KeyboardTypes keyboardType, KeyboardLayouts layoutType);
@@ -46,7 +46,7 @@ void gok_main_set_first_keyboard (GokKeyboard* pKeyboard);
 GokKeyboard* gok_main_get_current_keyboard (void);
 GtkWidget* gok_main_get_main_window (void);
 Accessible* gok_main_get_foreground_accessible (void);
-void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height);
+void gok_main_resize_window (GtkWidget* pWindow, GokKeyboard *pKeyboard, gint Width, gint Height);
 void gok_main_store_window_center (void);
 Accessible* gok_main_get_foreground_window_accessible (void);
 void gok_main_read_keyboards (void);
@@ -68,6 +68,8 @@ void gok_main_mousebutton_listener (gint button, gint state, long mods, long tim
 void gok_main_close (void);
 gboolean gok_main_get_use_geometry (void);
 void gok_main_get_geometry (GdkRectangle* pRectangle);
+void gok_main_update_struts (gint width, gint height);
+void gok_main_set_wm_dock (gboolean is_dock);
 
 #ifdef __cplusplus
 }
